#include "aldor"
#include "aldorio"

#library G "graph.ao"
import from G;
inline from G;

-- Rather horrid makefile generation
-- Problems with this lot:
-- 1. Too much list searching - set's representation is a problem here
-- 2. No way for a type to export a view of itself as a different type - cf java's 'Map.keySet', etc.
-- 3. Needs to be more data driven - all we're doing is building a dependency graph from .dep files

local generateMakefile(args: Array String): () == {
      import from MachineInteger;
      import from List String;
      import from String;


      stdout << args << newline;
      topTypesArg: List String := readNames(args.0);
      files:       List String := readNames(args.1);
      depsPairs:   List String := readNames(args.2);

      fileMap: HashTable(String, List String) := table();
      allTypes: Set String := [];

      fileMap.("base")      := [];
      fileMap.("axiom")     := [];
      fileMap.("initaxiom") := [];
      fileMap.("aldorext")  := [];

      for (class, id) in extractPairs files repeat {
          fileMap(class) := cons(id, fileMap(class));
          allTypes := union!(allTypes, id);
      }     
      
      stdout << "AllTypes: " << allTypes << newline;

      topTypes: List String := [];

      topTypes := [arg for arg in topTypesArg];
      for id in fileMap("aldorext") repeat {
          topTypes := cons(id, topTypes);
          topTypes := append!(topTypes, readNames("deps/" + id + ".dep"));
      }

      topLevelAncestors: Set String := findTopLevelConstructors(fileMap("axiom"), topTypes);
      stdout << "TopLevel: " << topLevelAncestors << newline;

      -- fixme: shouldn't these be defined elsewhere?
      topLevelAncestors := union!(topLevelAncestors, "TUPLE");
      topLevelAncestors := union!(topLevelAncestors, "EXIT");
      for i in topLevelAncestors repeat stdout << "top " << i << newline;
      allNodes: List String := [];
      allNodes := append!(allNodes, [id for id in allTypes]);

      stdout << "All Nodes: " << allNodes << newline;
      g: B__Graph(String) := graph(allNodes);
      for s in fileMap("axiom") repeat 
          for baseType in fileMap("base") repeat 
              addLink!(g, baseType, s);

      for s in fileMap("initaxiom") repeat {
          addLink!(g, s, substring(s, 5));
          for baseType in fileMap("base") repeat 
              addLink!(g, baseType, s);
      }

      for s in fileMap("aldorext") repeat 
          for baseType in fileMap("base") repeat 
              addLink!(g, baseType, s);
            
     stdout << "closure dependencies" << newline;
     
     for s in union([i for i in fileMap."initaxiom"], topLevelAncestors) repeat {
         depIds: Set String := [id for id in readNames("deps/" + s + ".dep")];

         for id in depIds repeat {
             if member?(("init__" + id), allTypes) then
                addLink!(g, "init__" + id, s);
             else 
                addLink!(g, id, s);
         }
     }

     stdout << "aldor dependencies" << newline;
     import from String;
     for id in fileMap("aldorext") repeat {
         for j in readNames("deps/" + id + ".dep") repeat {
             addLink!(g, j, id);
         }
     }

     upperTypes: Set String := [s for s in allTypes];
     upperTypes := [x for x in upperTypes | not member?(x, topLevelAncestors)];
     upperTypes := [x for x in upperTypes | not member?(x, fileMap.("base"))];
     upperTypes := [x for x in upperTypes | not member?(x, fileMap.("initaxiom"))];

     stdout << "upper types" << upperTypes << newline;

     stdout << "top level dependencies" << newline;
     for id in upperTypes repeat {
         for dep in readNames("deps/" + id + ".dep") repeat {
             addLink!(g, dep, id);
         }
     }


     for id in upperTypes repeat {
         if not member?(id, fileMap."aldorext") then 
            for alid in fileMap."aldorext" repeat
                addLink!(g, alid, id);

         if member?("init__"+id, fileMap."initaxiom") then 
             addLink!(g, "init__"+id, id);

         for dep in readNames("deps/" + id + ".dep") repeat
             addLink!(g, dep, id);
     }


     -- Finally, add the extra dependencies
     for (id,dep) in extractPairs depsPairs repeat {
         addLink!(g, dep, id);
     }

     import from GraphAlgorithms(String, B__Graph);
     s: Set Set String := connectedComponents g;

     stdout << "ConnectedComponents: " << s << newline;
     
     writeComponents("axiom", g, s);
}

local writeComponents(dname: String, g: B__Graph String, s: Set Set String): () == {
      import from Set String;
      import from Edge String;
      f: File := open(dname + "/spadset.mk", fileWrite$File);
      out: TextWriter := coerce(f);

      out << ("SPADSET__" + dname + " := ") << newline;
      idToSetMap: HashTable(String, String) := table();
      for clq in s repeat
          for id in clq repeat idToSetMap.id := firstName clq;

      for clq in s repeat {
          clqid := firstName(clq);
          out << ("SPADSET__" + dname + " += " + clqid )<< newline;
          out << ("SPADSET__ITEM__" + dname + "__" + clqid + " := \" ) << newline;
          for id in clq repeat {
              out << "  " + id + " \" << newline;
          }
          out << newline;
          
          out << ("SPADSET__DEPS__" + dname + "__" + clqid + " := \" ) << newline;
          deps: Set String := [ idToSetMap.(p0 e) for e in (flatten(Edge String, inEdges(g, id) for id in clq)) | not member?(p0 e, clq)];
          for id in deps repeat {
              out << "  " + id + " \" << newline;
          }
          out << newline;
      }
      flush! out;
      close! f;
}

local flatten(T: with, g: Generator List T): Generator T == {
      import from List T;
  generate {
      for l in g repeat for i in l repeat yield i
}
}


local findTopLevelConstructors(types: List(String), topTypes: List(String) ): Set(String) == {
        g: B__Graph String := graph(types);

	topDeps: Set(String) := [];
	for id: String in topTypes repeat {
	    topDeps := union!(topDeps, [x for x in readNames("deps/" + id + ".dep")]);
	}

        for id in types repeat {
            file: String := ("deps/" + id + ".dep");
            deps: List String := readNames(file);
            addDependencies(g, id, deps);
        }
        
        topLevelAncestors: Set String := [];
        for id in topDeps repeat {
            a: Set(String)  := ancestors(g, id);
	    topLevelAncestors := union!(topLevelAncestors, id);
            topLevelAncestors := union!(topLevelAncestors, a);
        }
	
	return topLevelAncestors;
}


local addDependencies(g: B__Graph String, id: String, l: List String): () == {
      import from MachineInteger;
      import from Set String;
      for dest in l repeat {
            if (not node?(dest, g)) then {
                stdout << "Missing node: " << dest << newline;
            }
            else 
                addLink!(g, dest, id);
        }
    }


local firstName(s: Set String): String == {
   import from Integer;
   local least: String;

   empty? s => never;
   for x in s for i in 1..1 repeat least := x;
   for x in s repeat if x < least then least := x;

   return least;
}

local extractPairs(l: List String): Generator Cross(String, String) ==
      generate {
          for str in l repeat {
	      r: TextReader := coerce str;
	      sb: StringBuffer := new();
	      readUntil!(sb, r, space);
	      s1: String := string(sb);
	      sb := new();
	      readWhile!(sb, r, (c: Character): Boolean +->  { not(space? c) and not (c=newline)});
	      s2: String := string(sb);
	      yield (s1, s2);
	  }
}

local extractPairs(l: Generator String): Generator Cross(String, String) ==
      generate {
          for str in l repeat {
	      r: TextReader := coerce str;
	      sb: StringBuffer := new();
	      readUntil!(sb, r, space);
	      s1: String := string(sb);
	      sb := new();
	      readWhile!(sb, r, (c: Character): Boolean +->  { not(space? c) and not (c=newline)});
	      s2: String := string(sb);
	      yield (s1, s2);
	  }
}

local readNames(fname: String): List String == {
    import from MachineInteger;
    local sb: StringBuffer;

    f: File := safeOpen(fname);
    rdr: TextReader := coerce f;

    ids: List String := [];
    sb := new();
    while (readUntil!( sb, rdr, newline) = newline) repeat {
         stdout<<sb<<newline; -- bug: this seems to be needed to avoid a compile bug
         ids := cons(string(sb), ids);
	 sb := new();
    }
    close! f;
stdout << fname << ": " << ids << newline;
    return ids;
}


local safeOpen(fname: String): File == {
      try { return open(fname); }
      catch E in {
              stdout << "Missing: " << fname << newline;
              never;
      }
}


main(): () == {
	generateMakefile(arguments$CommandLine);
}

stdout << "Hello" << newline;
main();




